import logging
import os
import sys

from container_app_conf.formatter.toml import TomlFormatter
from prometheus_client import start_http_server

from grocy_telegram_bot.bot import GrocyTelegramBot
from grocy_telegram_bot.config import Config

parent_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), "..", ".."))
sys.path.append(parent_dir)

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
LOGGER = logging.getLogger(__name__)


def main():
    config = Config()

    log_level = logging._nameToLevel.get(str(config.LOG_LEVEL.value).upper(), config.LOG_LEVEL.default)
    logging.getLogger("grocy_telegram_bot").setLevel(log_level)

    LOGGER.debug("Config:\n{}".format(config.print(TomlFormatter())))

    # start prometheus server
    if config.STATS_ENABLED.value:
        start_http_server(config.STATS_PORT.value)

    grocy_telegram_bot = GrocyTelegramBot(config)
    grocy_telegram_bot.start()


if __name__ == '__main__':
    main()



from datetime import datetime

from telegram import ParseMode, Update
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.argument import Flag
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_CHORES
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.stats import COMMAND_TIME_CHORES
from grocy_telegram_bot.util import send_message, filter_overdue_chores, chore_to_str


class ChoreCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_CHORES,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._chores_callback)
        ]

    @command(
        name=COMMAND_CHORES,
        description="List overdue chores.",
        arguments=[
            Flag(name=["all", "a"], description="Show all chores")
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_CHORES.time()
    def _chores_callback(self, update: Update, context: CallbackContext, all: bool) -> None:
        """
        Show a list of all chores
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id

        chores = self._grocy.chores(True)
        chores = sorted(
            chores,
            key=lambda
                x: datetime.now().astimezone() if x.next_estimated_execution_time is None else x.next_estimated_execution_time
        )

        overdue_chores = filter_overdue_chores(chores)
        other = [item for item in chores if item not in overdue_chores]

        overdue_item_texts = list(map(chore_to_str, overdue_chores))
        other_item_texts = list(map(chore_to_str, other))

        lines = ["*=> Chores <=*"]
        if all and len(other_item_texts) > 0:
            lines.extend([
                "",
                *other_item_texts
            ])

        if len(overdue_item_texts) > 0:
            lines.extend([
                "",
                "*Overdue:*",
                *overdue_item_texts
            ])

        text = "\n".join(lines).strip()
        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN)



from typing import List, Any

from pygrocy import Grocy

from grocy_telegram_bot.monitoring.watcher import GrocyEntityWatcher
from grocy_telegram_bot.stats import TASK_WATCHER_TIME


class TaskWatcher(GrocyEntityWatcher):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(grocy, on_update_listener, interval)

    def _fetch_data(self) -> List[Any]:
        # TODO: pygrocy doesn't support tasks yet
        return []

    @TASK_WATCHER_TIME.time()
    def _run(self):
        super()._run()



from grocy_telegram_bot.util import fuzzy_match
from tests import TestBase


class FuzzyMatchTest(TestBase):

    def test_fuzzy_match(self):
        term = "new york jets"
        choices = ["Atlanta Falcons", "New York Jets", "New York Giants", "Dallas Cowboys"]

        matches = fuzzy_match(term, choices)

        self.assertIsNotNone(matches)



import logging
from typing import Dict

from telegram import Update, InlineKeyboardMarkup, InlineKeyboardButton
from telegram.ext import CallbackContext


class InlineKeyboardHandler:
    # this map is used to map a command_id to a callback function
    _inline_keyboard__command_to_callback_map = {}

    # this map is used to remember what type is associated
    # with the callback data of a keyboard button
    _inline_keyboard__callback_data_type_map = {}

    def __init__(self):
        pass

    def register_listener(self, chat_id: str, message_id: str, command_id: str, callback, callback_data: any):
        key = f"{chat_id}_{message_id}"
        self._inline_keyboard__callback_data_type_map[key] = command_id
        self._inline_keyboard__command_to_callback_map[command_id] = {
            "callback": callback,
            "callback_data": callback_data
        }

    def inline_keyboard_click_callback(self, update: Update, context: CallbackContext):
        """
        Handles inline keyboard button click callbacks
        :param update:
        :param context:
        """
        bot = context.bot
        user_id = update.callback_query.from_user.id
        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id

        query = update.callback_query
        # TODO: is it possible to get this query_id when initially sending the keyboard? -> NOPE, but a chat_id & msg_id key should suffice
        # then it would be easy to remember the shopping list items in a dict somewhere
        # to regenerate it with the new counts/amounts
        query_id = query.id
        selection_data = query.data

        try:
            command_id = self._inline_keyboard__callback_data_type_map.get(f"{chat_id}_{message_id}", None)
            if command_id is None:
                bot.answer_callback_query(query_id, text="Unknown message")
                return

            callback_info = self._inline_keyboard__command_to_callback_map[command_id]
            callback = callback_info["callback"]
            callback_data = callback_info["callback_data"]
            # call listener
            callback(update, context, selection_data, callback_data)
        except Exception:
            logging.exception("Error processing inline keyboard button callback")
            bot.answer_callback_query(query_id, text="Error")

    @staticmethod
    def build_inline_keyboard(items: Dict[str, str]) -> InlineKeyboardMarkup:
        """
        Builds an inline button menu
        :param items: dictionary of "button text" -> "callback data" items
        :return: reply markup
        """
        keyboard = list(map(lambda x: InlineKeyboardButton(x[0], callback_data=x[1]), items.items()))
        return InlineKeyboardMarkup.from_column(keyboard)



import functools
import hashlib
import logging
import operator
import os
from datetime import datetime, timezone, timedelta
from functools import wraps
from io import BytesIO
from time import time
from typing import List, Any, Tuple

from pygrocy.grocy import Chore, Product, ShoppingListProduct
from telegram import Bot, Message, ReplyMarkup

from grocy_telegram_bot.config import Config
from grocy_telegram_bot.const import TELEGRAM_CAPTION_LENGTH_LIMIT, NEVER_EXPIRES_DATE

LOGGER = logging.getLogger(__name__)

CONFIG = Config()


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        td = te - ts
        LOGGER.debug('func:%r took: %2.4f sec' % (f.__name__, td))
        return result

    return wrap


# def download_image_bytes(url: str) -> bytes:
#     """
#     Downloads the image from the given url
#     :return: the downloaded image
#     """
#     image = requests.get(url, timeout=REQUESTS_TIMEOUT)
#     image.raise_for_status()
#     return image.content

def flatten(data: List[List[Any]]) -> List[Any]:
    """
    Flattens a list of lists
    :param data: the data to flatten
    :return: flattened list
    """
    return functools.reduce(operator.iconcat, data, [])


def create_hash(data: bytes) -> str:
    """
    Creates a hash of the given bytes
    :param data: data to hash
    :return: hash
    """
    return hashlib.md5(data).hexdigest()


def cryptographic_hash(data: bytes or str) -> str:
    """
    Creates a cryptographic hash of the given bytes
    :param data: data to hash
    :return: hash
    """
    if isinstance(data, str):
        data = data.encode()
    import hashlib
    hash = hashlib.sha512(data).hexdigest()
    return hash


def send_photo(bot: Bot, chat_id: str, file_id: int or None = None, image_data: bytes or None = None,
               caption: str = None) -> [str]:
    """
    Sends a photo to the given chat
    :param bot: the bot
    :param chat_id: the chat product_id to send the image to
    :param file_id: the telegram file product_id of the already uploaded image
    :param image_data: the image data
    :param caption: an optional image caption
    :return: a set of telegram image file_id's
    """
    if image_data is not None:
        image_bytes_io = BytesIO(image_data)
        image_bytes_io.name = 'image.jpeg'
        photo = image_bytes_io
    elif file_id is not None:
        photo = file_id
    else:
        raise ValueError("At least one of file_id and image_data has to be provided!")

    if caption is not None:
        caption = _format_caption(caption)

    message = bot.send_photo(chat_id=chat_id, photo=photo, caption=caption)
    return set(map(lambda x: x.file_id, message.photo))


def format_for_single_line_log(text: str) -> str:
    """
    Formats a text for log
    :param text:
    :return:
    """
    text = "" if text is None else text
    return " ".join(text.split())


def _format_caption(text: str) -> str or None:
    if text is None:
        return None

    # remove empty lines
    text = os.linesep.join([s for s in text.splitlines() if s.strip()])
    # limit to 200 characters (telegram api limitation)
    if len(text) > TELEGRAM_CAPTION_LENGTH_LIMIT:
        text = text[:197] + "…"

    return text


def datetime_fmt_date_only(d: datetime):
    from babel.dates import format_date
    time = d.astimezone()
    return format_date(time.date(), locale=CONFIG.LOCALE.value)


def send_message(bot: Bot, chat_id: str, message: str, parse_mode: str = None, reply_to: int = None,
                 menu: ReplyMarkup = None) -> Message:
    """
    Sends a text message to the given chat
    :param bot: the bot
    :param chat_id: the chat product_id to send the message to
    :param message: the message to chat (may contain emoji aliases)
    :param parse_mode: specify whether to parse the text as markdown or HTML
    :param reply_to: the message product_id to reply to
    :param menu: inline keyboard menu markup
    """
    from emoji import emojize

    emojized_text = emojize(message, use_aliases=True)
    return bot.send_message(chat_id=chat_id, parse_mode=parse_mode, text=emojized_text, reply_to_message_id=reply_to,
                            reply_markup=menu)


def product_to_str(item: Product) -> str:
    from pygrocy.utils import parse_int
    amount = parse_int(item.available_amount, item.available_amount)

    text = f"{amount}x\t{item.name}"
    if item.best_before_date.date() < NEVER_EXPIRES_DATE:
        expire_date = datetime_fmt_date_only(item.best_before_date)
        text += f" (Exp: {expire_date})"

    return text


def chore_to_str(chore: Chore) -> str:
    """
    Converts a chore object into a string representation
    :param chore: the chore item
    :return: a text representation
    """
    today_utc_date_with_zero_time = datetime.today().astimezone(tz=timezone.utc)

    days_off = None
    date_str = None
    if chore.next_estimated_execution_time is not None:
        days_off = abs((chore.next_estimated_execution_time - today_utc_date_with_zero_time).days)
        date_str = datetime_fmt_date_only(chore.next_estimated_execution_time)

    lines = [chore.name]
    if days_off is not None:
        lines.append(f"  Due: {days_off} days ({date_str})")

    return "\n".join(lines)


def shopping_list_item_to_str(item: ShoppingListProduct) -> str:
    """
    Converts a shopping list item object into a string representation
    :param item: the shopping list item
    :return: a text representation
    """
    from pygrocy.utils import parse_int
    amount = parse_int(item.amount, item.amount)

    return f"{amount}x {item.product.name}"


def filter_overdue_chores(chores: List[Chore]) -> List[Chore]:
    today_utc_date_with_zero_time = datetime.today().astimezone(tz=timezone.utc)
    return list(filter(lambda x:
                       x.next_estimated_execution_time is not None
                       and x.next_estimated_execution_time <= today_utc_date_with_zero_time, chores))


def filter_has_expiry_products(products: List[Product]):
    never_expires_date = datetime(year=2999, month=12, day=31).astimezone(tz=timezone.utc)
    return list(filter(lambda x: x.best_before_date < never_expires_date, products))


def filter_expiring_products(products: List[Product], days_to_expiry: int = 5):
    today_minus_expiry_timeframe = datetime.today().astimezone(tz=timezone.utc) - timedelta(days=days_to_expiry)
    products_with_expiry = filter_has_expiry_products(products)
    return list(filter(lambda x: x.best_before_date < today_minus_expiry_timeframe, products_with_expiry))


def filter_expired_products(products: List[Product]):
    date_today = datetime.today().astimezone(tz=timezone.utc)
    products_with_expiry = filter_has_expiry_products(products)
    return list(filter(lambda x: x.best_before_date < date_today, products_with_expiry))


def filter_new_by_key(a: List, b: List, key: callable) -> List:
    """
    Returns a list of all items, that are new in b when compared to a,
    using the key function to determine a unique identifier for list items
    :param a: "old" list
    :param b: "new" list
    :param key: function to map list items to a unique identifier
    :return: new list items
    """
    a_ids = set(map(key, a))
    b_ids = set(map(key, b))
    new_ids = b_ids - a_ids

    result = []
    for id in new_ids:
        item_in_b = list(filter(lambda x: key(x) == id, b))[0]
        result.append(item_in_b)
    return result


def fuzzy_match(term: str, choices: List[Any], limit: int = None, key=lambda x: x, ignorecase: bool = True) -> List[
    Tuple[Any, int]]:
    """
    Does a fuzzy search on the given
    :param term: the search term
    :param choices: list of possible choices
    :param key: function to turn a choice item into a string
    :param limit: Optional maximum for the number of elements returned
    :return: List of (choice, ratio) tuples, sorted by descending ratio
    """
    # map choices to key
    if ignorecase:
        term = term.casefold()
    choices = filter(lambda x: key(x) is not None, choices)
    key_map = dict(map(lambda x: (key(x).casefold() if ignorecase else key(x), x), choices))

    from fuzzywuzzy import process
    from fuzzywuzzy import fuzz
    matches = process.extract(term, key_map.keys(), limit=limit, scorer=fuzz.UWRatio)

    # map results back to original choices
    result = list(map(lambda x: (key_map[x[0]], x[1]), matches))

    return result



from telegram import Update, ParseMode
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.decorator import command
from telegram_click.permission import PRIVATE_CHAT

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_CONFIG
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.util import send_message


class ConfigCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_CONFIG,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._config_command_callback),
        ]

    @command(
        name=COMMAND_CONFIG,
        description="Show current application configuration.",
        permissions=PRIVATE_CHAT & CONFIG_ADMINS
    )
    def _config_command_callback(self, update: Update, context: CallbackContext):
        """
        /config command handler
        :param update: the chat update object
        :param context: telegram context
        """
        from container_app_conf.formatter.toml import TomlFormatter

        bot = context.bot
        chat_id = update.effective_message.chat_id
        message_id = update.effective_message.message_id

        text = self._config.print(formatter=TomlFormatter())
        text = "```\n{}\n```".format(text)
        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN, reply_to=message_id)



import logging
from datetime import timedelta

from telegram import Update, ParseMode
from telegram.ext import CommandHandler, Filters, MessageHandler, Updater, \
    CallbackContext, CallbackQueryHandler
from telegram_click.decorator import command

from grocy_telegram_bot.bot.inline_keyboard_handler import InlineKeyboardHandler
from grocy_telegram_bot.bot.reply_keyboard_handler import ReplyKeyboardHandler
from grocy_telegram_bot.cache import GrocyCached
from grocy_telegram_bot.commands.chore import ChoreCommandHandler
from grocy_telegram_bot.commands.config import ConfigCommandHandler
from grocy_telegram_bot.commands.help import HelpCommandHandler
from grocy_telegram_bot.commands.inventory import InventoryCommandHandler
from grocy_telegram_bot.commands.shopping_list import ShoppingListCommandHandler
from grocy_telegram_bot.commands.stats import StatsCommandHandler
from grocy_telegram_bot.commands.version import VersionCommandHandler
from grocy_telegram_bot.config import Config
from grocy_telegram_bot.const import *
from grocy_telegram_bot.monitoring.monitor import Monitor
from grocy_telegram_bot.notifier import Notifier
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.stats import COMMAND_TIME_START
from grocy_telegram_bot.util import send_message, flatten

LOGGER = logging.getLogger(__name__)


class GrocyTelegramBot:
    """
    The main entry class of the grocy telegram bot
    """

    def __init__(self, config: Config):
        """
        Creates an instance.
        :param config: configuration object
        """
        self._config = config

        self._grocy = GrocyCached(
            base_url=f"{config.GROCY_HOST.value}",
            api_key=config.GROCY_API_KEY.value,
            port=config.GROCY_PORT.value)

        self._response_handler = ReplyKeyboardHandler()
        self._keyboard_handler = InlineKeyboardHandler()

        self._updater = Updater(token=self._config.TELEGRAM_BOT_TOKEN.value, use_context=True)
        LOGGER.debug("Using bot id '{}' ({})".format(self._updater.bot.id, self._updater.bot.name))

        self._dispatcher = self._updater.dispatcher

        self._help_command_handler = HelpCommandHandler(self._config, self._grocy, self._response_handler,
                                                        self._keyboard_handler)
        self._grocy_command_handlers = [
            ChoreCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler),
            ConfigCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler),
            self._help_command_handler,
            InventoryCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler),
            ShoppingListCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler),
            StatsCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler),
            VersionCommandHandler(self._config, self._grocy, self._response_handler, self._keyboard_handler)
        ]

        command_handlers = flatten(list(map(lambda x: x.command_handlers(), self._grocy_command_handlers)))

        handler_groups = {
            0: [CallbackQueryHandler(callback=self._keyboard_handler.inline_keyboard_click_callback)],
            1: command_handlers +
               [
                   CommandHandler(COMMAND_START,
                                  filters=(~ Filters.reply) & (~ Filters.forwarded),
                                  callback=self._start_callback),
                   CommandHandler(COMMAND_CHAT_ID,
                                  filters=(~ Filters.reply) & (~ Filters.forwarded),
                                  callback=self._chat_id_callback),

                   CommandHandler(CANCEL_KEYBOARD_COMMAND[1:],
                                  filters=(~ Filters.reply) & (~ Filters.forwarded),
                                  callback=self._response_handler.cancel_keyboard_callback),
                   # unknown command handler
                   MessageHandler(
                       filters=Filters.command & (~ Filters.forwarded),
                       callback=self._unknown_command_callback),
                   MessageHandler(
                       filters=(~ Filters.forwarded),
                       callback=self._any_message_callback),
               ]
        }

        for group, handlers in handler_groups.items():
            for handler in handlers:
                self._updater.dispatcher.add_handler(handler, group=group)

        self._monitor = None
        chat_ids = self._config.NOTIFICATION_CHAT_IDS.value
        if chat_ids is not None and len(chat_ids) > 0:
            self._notifier = Notifier(self._updater, chat_ids)
            interval = self._config.GROCY_CACHE_DURATION.value + timedelta(seconds=1)
            self._monitor = Monitor(interval, self._notifier, self._grocy)

    @property
    def bot(self):
        return self._updater.bot

    def start(self):
        """
        Starts up the bot.
        """
        if self._monitor is not None:
            self._monitor.start()
        self._updater.start_polling()
        self._updater.idle()

    def stop(self):
        """
        Shuts down the bot.
        """
        if self._monitor is not None:
            self._monitor.stop()
        self._updater.stop()

    @COMMAND_TIME_START.time()
    def _start_callback(self, update: Update, context: CallbackContext) -> None:
        """
        Welcomes a new user with a greeting message
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        user_first_name = update.effective_user.first_name

        if not CONFIG_ADMINS.evaluate(update, context):
            send_message(bot, chat_id, "Sorry, you do not have permissions to use this bot.")
            return

        send_message(bot, chat_id,
                     f"Welcome {user_first_name},\nthis is your grocy-telegram-bot instance, ready to go!")

    @command(
        name=COMMAND_CHAT_ID,
        description="Show current chat id.",
        permissions=CONFIG_ADMINS
    )
    def _chat_id_callback(self, update: Update, context: CallbackContext) -> None:
        """
        Print the current chat id, to make it easier to add it to notifications
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        send_message(bot, chat_id, f"{chat_id}", parse_mode=ParseMode.MARKDOWN)

    def _unknown_command_callback(self, update: Update, context: CallbackContext) -> None:
        """
        Handles unknown commands send by a user
        :param update: the chat update object
        :param context: telegram context
        """
        message = update.effective_message
        username = "N/A"
        if update.effective_user is not None:
            username = update.effective_user.username

        user_is_admin = username in self._config.TELEGRAM_ADMIN_USERNAMES.value
        if user_is_admin:
            self._help_command_handler.help_command_callback(update, context)
            return

    def _any_message_callback(self, update: Update, context: CallbackContext) -> None:
        """
        Used to respond to response keyboard entry selections
        :param update: the chat update object
        :param context: telegram context
        """
        self._response_handler.on_message(update, context)



from pygrocy.grocy import Chore
from pygrocy.grocy_api_client import CurrentChoreResponse

from grocy_telegram_bot.util import filter_new_by_key
from tests import TestBase


class MonitorTest(TestBase):

    def test_filter_by_new_key_single_new(self):
        old = [
            Chore(CurrentChoreResponse(self._generate_chore_json(1))),
            Chore(CurrentChoreResponse(self._generate_chore_json(2))),
            Chore(CurrentChoreResponse(self._generate_chore_json(3)))
        ]

        new = [
            Chore(CurrentChoreResponse(self._generate_chore_json(4)))
        ]
        combined = old + new

        filtered = filter_new_by_key(old, combined, key=lambda x: x.id)

        self.assertEquals(len(filtered), 1)
        self.assertIn(new[0], filtered)

    def test_filter_by_new_key_new_and_removed_old(self):
        old = [
            Chore(CurrentChoreResponse(self._generate_chore_json(1))),
            Chore(CurrentChoreResponse(self._generate_chore_json(2))),
            Chore(CurrentChoreResponse(self._generate_chore_json(3)))
        ]

        new = [
            Chore(CurrentChoreResponse(self._generate_chore_json(1))),
            Chore(CurrentChoreResponse(self._generate_chore_json(2))),
            Chore(CurrentChoreResponse(self._generate_chore_json(4)))
        ]

        filtered = filter_new_by_key(old, new, key=lambda x: x.id)

        self.assertEquals(len(filtered), 1)
        self.assertIn(new[2], filtered)

    @staticmethod
    def _generate_chore_json(id: int) -> dict:
        return {
            "chore_id": id,
            "last_tracked_time": None,
            "next_estimated_execution_time": None
        }



from collections import OrderedDict
from typing import Tuple, List, Dict

from pygrocy.grocy import ShoppingListProduct, Product
from telegram import Update, ParseMode, ReplyKeyboardRemove
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.argument import Argument, Flag
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_SHOPPING_LIST, COMMAND_SHOPPING, NEVER_EXPIRES_DATE, \
    COMMAND_SHOPPING_LIST_ADD
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.stats import COMMAND_TIME_SHOPPING_LIST, COMMAND_TIME_SHOPPING, COMMAND_TIME_SHOPPING_LIST_ADD
from grocy_telegram_bot.telegram_util import ShoppingListItemButtonCallbackData
from grocy_telegram_bot.util import send_message, shopping_list_item_to_str


class ShoppingListCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_SHOPPING,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._shopping_callback),
            CommandHandler(COMMAND_SHOPPING_LIST,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._shopping_lists_callback),
            CommandHandler(COMMAND_SHOPPING_LIST_ADD,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._shopping_list_add_callback),
        ]

    @command(
        name=COMMAND_SHOPPING_LIST_ADD,
        description="Add an item to a shopping list.",
        arguments=[
            Argument(name=["name"], description="Product name", example="Banana"),
            Argument(name=["amount"], description="Product amount", type=int, example="2",
                     validator=lambda x: x > 0, optional=True, default=1),
            Argument(name=["id"], description="Shopping list id", type=int, example="1",
                     optional=True, default=1),
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_SHOPPING_LIST_ADD.time()
    def _shopping_list_add_callback(self, update: Update, context: CallbackContext,
                                    name: str, amount: int, id: int) -> None:
        """
        Show a list of all shopping lists
        :param update: the chat update object
        :param context: telegram context
        :param name: product name
        :param amount: product amount
        :param id: shopping list id
        """

        products = self._grocy.get_all_products()
        self._reply_keyboard_handler.await_user_selection(
            update, context, name, choices=products, key=lambda x: x.name,
            callback=self._add_product_keyboard_response_callback,
            callback_data={
                "product_name": name,
                "amount": amount,
                "shopping_list_id": id
            }
        )

    def _add_product_keyboard_response_callback(self, update: Update, context: CallbackContext,
                                                product: Product, data: dict):
        """
        Called when the user has selected a product to add to a shopping list
        :param update: the chat update object
        :param context: telegram context
        :param product: the selected product
        :param data: callback data
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id

        amount = data["amount"]
        shopping_list_id = data["shopping_list_id"]
        self._grocy.add_product_to_shopping_list(product.id, shopping_list_id, amount)

        text = "Added {}x {}".format(amount, product.name)
        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN, reply_to=message_id,
                     menu=ReplyKeyboardRemove(selective=True))

    @command(
        name=COMMAND_SHOPPING,
        description="Print shopping list with buttons to check off items.",
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_SHOPPING.time()
    def _shopping_callback(self, update: Update, context: CallbackContext) -> None:
        """
        Show a list of all shopping lists
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id

        shopping_list_items = self._grocy.shopping_list(True)
        # TODO: sort by parent product / product category
        # TODO: let the user specify the order of product categories, according to the grocery store of his choice
        shopping_list_items = sorted(shopping_list_items, key=lambda x: x.product.name.lower())

        # generate message
        text = "*=> Shopping List <=*"
        # generate keyboard
        initial_button_tuples = self._create_shopping_list_item_button_tuples(shopping_list_items)
        inline_keyboard_items = self._create_shopping_list_keyboard_items(initial_button_tuples)
        inline_keyboard_markup = self._inline_keyboard_handler.build_inline_keyboard(inline_keyboard_items)

        # send message
        result = send_message(bot, chat_id, text, menu=inline_keyboard_markup, parse_mode=ParseMode.MARKDOWN)
        # register callback for button presses
        self._inline_keyboard_handler.register_listener(
            chat_id=chat_id,
            message_id=result.message_id,
            command_id=ShoppingListItemButtonCallbackData.command_id,
            callback=self._shopping_button_pressed_callback,
            callback_data={
                "shopping_list_items": shopping_list_items,
                "initial_keyboard_items": initial_button_tuples
            }
        )

    def _shopping_button_pressed_callback(self, update: Update, context: CallbackContext, button_data: str, data: Dict):
        button_data = ShoppingListItemButtonCallbackData.parse(button_data)
        query = update.callback_query
        query_id = query.id

        # TODO: Should this be a configuration option or a flag?
        #  if a flag should be used, this would need to be saved for every keyboard invididually...
        remove_finished_products_from_keyboard = True

        # TODO: there is currently no way to query shopping list ids, so this is hardcoded for now
        shopping_list_id = 1

        # retrieve shopping list from grocy
        # TODO: use list in store or from api?
        # shopping_list_items = self._grocy.shopping_list(True)
        shopping_list_items = data["shopping_list_items"]
        # find the matching item
        matching_items = list(filter(lambda x: x.id == button_data.shopping_list_item_id, shopping_list_items))
        if len(matching_items) <= 0:
            # if the item is not on the shopping list anymore, show a message
            context.bot.answer_callback_query(query_id, text=f"The item is not on the shopping list anymore.")
            # TODO: it should still be possible to add items beyond what the shopping list has
            #  if "remove_finished_products_from_keyboard" is False
            return

        # if the item is still on the shopping list, remove one item
        item = matching_items[0]
        response = self._grocy.remove_product_in_shopping_list(item.product_id, shopping_list_id, amount=1)
        if response is not None:
            response.raise_for_status()
        # also remove it from the keyboard
        stored_button_tuples = data["initial_keyboard_items"]
        # find the item in button tuples
        stored_item_title, stored_item_data = \
            list(filter(lambda x: x[1].shopping_list_item_id == item.id, stored_button_tuples.items()))[0]
        # increase its "clicked" counter
        stored_item_data.button_click_count += 1
        # generate the new button text
        stored_item_title_new = self._generate_button_title(item, stored_item_data)
        # remove the old key
        stored_button_tuples.pop(stored_item_title)
        if stored_item_data.button_click_count >= stored_item_data.shopping_list_amount and remove_finished_products_from_keyboard:
            # dont put the data back
            pass
        else:
            # put the modified data back in the dictionary
            stored_button_tuples[stored_item_title_new] = stored_item_data

        # add one item to the inventory
        product = item.product
        response = self._grocy.add_product(product_id=product.id, amount=1, price=None,
                                           best_before_date=NEVER_EXPIRES_DATE)
        if response is not None:
            response.raise_for_status()

        # regenerate keyboard
        keyboard_items = self._create_shopping_list_keyboard_items(stored_button_tuples)
        # we need to re-sort since the items have changed
        keyboard_items = OrderedDict(sorted(keyboard_items.items(), key=lambda x: x[0].lower()))
        inline_keyboard_markup = self._inline_keyboard_handler.build_inline_keyboard(keyboard_items)

        query.edit_message_reply_markup(reply_markup=inline_keyboard_markup)
        answer_text = f"Checked off '{product.name}' ({stored_item_data.button_click_count}/{stored_item_data.shopping_list_amount})"
        context.bot.answer_callback_query(query_id, text=answer_text)

    @command(
        name=COMMAND_SHOPPING_LIST,
        description="List shopping list items.",
        arguments=[
            Argument(name=["id"], description="Shopping list id", type=int, example="1",
                     optional=True, default=1),
            Flag(name=["add_missing", "a"],
                 description="Add items below minimum stock to the shopping list")
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_SHOPPING_LIST.time()
    def _shopping_lists_callback(self, update: Update, context: CallbackContext,
                                 id: int, add_missing: bool or None) -> None:
        """
        Show a list of all shopping lists
        :param update: the chat update object
        :param context: telegram context
        :param id: shopping list id
        :param add_missing: whether to add missing products to the shopping list before displaying
        """
        bot = context.bot
        chat_id = update.effective_chat.id

        if add_missing:
            self._grocy.add_missing_product_to_shopping_list(shopping_list_id=id)

        # TODO: when supported, pass shopping list id here
        shopping_list_items = self._grocy.shopping_list(True)
        shopping_list_items = sorted(shopping_list_items, key=lambda x: x.product.name.lower())

        item_texts = list(list(map(shopping_list_item_to_str, shopping_list_items)))
        text = "\n".join([
            "*=> Shopping List <=*",
            *item_texts,
        ]).strip()

        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN)

    def _create_shopping_list_keyboard_items(self, items: Dict[str, ShoppingListItemButtonCallbackData]) -> Dict[
        str, str]:
        return OrderedDict(list(map(lambda x: (x[0], x[1].minify()), items.items())))

    def _create_shopping_list_item_button_tuples(self, items: List[ShoppingListProduct]) -> Dict[
        str, ShoppingListItemButtonCallbackData]:
        """
        Creates the data required for generating a keyboard button from shopping list items
        :param items: shopping list items
        :return : button title, button callback data
        """
        return OrderedDict(list(map(self._create_shopping_list_item_button_tuple, items)))

    def _create_shopping_list_item_button_tuple(self, item: ShoppingListProduct) -> Tuple[
        str, ShoppingListItemButtonCallbackData]:
        """
        Creates the data required for generating a keyboard button from a shopping list item
        :param item: shopping list item
        :return : button title, button callback data
        """
        button_data = self._create_shopping_list_item_button_data(item)
        button_title = self._generate_button_title(item, button_data)
        return button_title, button_data

    @staticmethod
    def _generate_button_title(item, stored_item_data) -> str:
        return f"{item.product.name} ({stored_item_data.button_click_count}/{int(item.amount)})"

    @staticmethod
    def _create_shopping_list_item_button_data(item: ShoppingListProduct,
                                               count: int = 0) -> ShoppingListItemButtonCallbackData:
        return ShoppingListItemButtonCallbackData(
            shopping_list_item_id=item.id,
            button_click_count=count,
            shopping_list_amount=int(item.amount)
        )



from telegram import Update
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_STATS
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.stats import format_metrics
from grocy_telegram_bot.util import send_message


class StatsCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_STATS,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._stats_callback),
        ]

    @command(
        name=COMMAND_STATS,
        description="List statistics of this bot.",
        permissions=CONFIG_ADMINS
    )
    def _stats_callback(self, update: Update, context: CallbackContext) -> None:
        """
        /stats command handler
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        text = format_metrics()

        send_message(bot, chat_id, text, reply_to=message.message_id)



from datetime import timedelta
from typing import List, Any

from pygrocy import Grocy
from pygrocy.grocy import Chore, Product
from pygrocy.grocy_api_client import ShoppingListItem

from grocy_telegram_bot.monitoring.watcher.chore import ChoreWatcher
from grocy_telegram_bot.monitoring.watcher.inventory import StockWatcher, VolatileStockWatcher
from grocy_telegram_bot.monitoring.watcher.shopping_list import ShoppingListWatcher
from grocy_telegram_bot.monitoring.watcher.task import TaskWatcher
from grocy_telegram_bot.notifier import Notifier
from grocy_telegram_bot.stats import TOTAL_CHORES_COUNT, OVERDUE_CHORES_COUNT, PRODUCT_INVENTORY_COUNT, \
    EXPIRED_PRODUCTS_COUNT, SHOPPING_LIST_ITEM_COUNT, TASK_COUNT
from grocy_telegram_bot.util import product_to_str, chore_to_str, filter_overdue_chores, filter_expired_products, \
    filter_expiring_products, filter_new_by_key


class Monitor:

    def __init__(self, interval: timedelta, notifier: Notifier, grocy: Grocy):
        self._notifier = notifier
        self._grocy = grocy

        interval_seconds = interval.total_seconds()

        self.watchers = [
            ChoreWatcher(self._grocy, self.on_chore_update, interval_seconds),
            StockWatcher(self._grocy, self.on_stock_update, interval_seconds),
            VolatileStockWatcher(self._grocy, self.on_volatile_stock_update, interval_seconds),
            ShoppingListWatcher(self._grocy, self.on_shopping_list_update, interval_seconds),
            TaskWatcher(self._grocy, self.on_task_update, interval_seconds)
        ]

    def start(self):
        """
        Start monitoring the state of grocy.
        """
        for watcher in self.watchers:
            watcher.start()

    def stop(self):
        """
        Stop monitoring the state of grocy.
        """
        for watcher in self.watchers:
            watcher.stop()

    def on_chore_update(self, old: List[Chore], new: List[Chore]):
        TOTAL_CHORES_COUNT.set(len(new))

        new_overdue = filter_overdue_chores(new)
        OVERDUE_CHORES_COUNT.set(len(new_overdue))

        if old is not None:
            old_overdue = filter_overdue_chores(old)
            self._notify_about_new_overdue_chores(old_overdue, new_overdue)

    def _notify_about_new_overdue_chores(self, old: List[Chore], new: List[Chore]):
        # check if a new chore is due
        new_overdue = filter_new_by_key(old, new, key=lambda x: x.id)

        lines = list(map(chore_to_str, new_overdue))
        # send notification if required
        if len(lines) > 0:
            message = "\n".join([
                "Chore(s) overdue:",
                *lines
            ])
            self._notifier.notify(message)

    def on_stock_update(self, old: List[Product], new: List[Product]):
        for product in new:
            PRODUCT_INVENTORY_COUNT.labels(product_name=product.name).set(product.available_amount)

    def on_volatile_stock_update(self, old: List[Product], new: List[Product]):
        for product in new:
            PRODUCT_INVENTORY_COUNT.labels(product_name=product.name).set(product.available_amount)

        new_expired = filter_expired_products(new)
        EXPIRED_PRODUCTS_COUNT.set(len(new_expired))

        if old is not None:
            old_expired = filter_expired_products(old)
            self._notify_about_new_expired_products(old_expired, new_expired)

            old_expiring = filter_expiring_products(old)
            new_expiring = filter_expiring_products(new)
            self._notify_about_new_expired_products(old_expiring, new_expiring)

    def _notify_about_new_expiring_products(self, old: List[Product], new: List[Product]):
        new_expiring = filter_new_by_key(old, new, key=lambda x: x.id)

        lines = list(map(product_to_str, new_expiring))
        if len(lines) > 0:
            message = "\n".join([
                "Product(s) expiring soon:",
                *lines
            ])
            self._notifier.notify(message)

    def _notify_about_new_expired_products(self, old: List[Product], new: List[Product]):
        new_expired = filter_new_by_key(old, new, key=lambda x: x.id)

        lines = list(map(product_to_str, new_expired))
        if len(lines) > 0:
            message = "\n".join([
                "Product(s) expired:",
                *lines
            ])
            self._notifier.notify(message)

    def on_shopping_list_update(self, old: List[ShoppingListItem], new: List[ShoppingListItem]):
        # TODO: when pygrocy supports multiple shopping lists, this has to be updated
        SHOPPING_LIST_ITEM_COUNT.labels(name="Shopping List").set(len(new))

    def on_task_update(self, old: List[Any], new: List[Any]):
        TASK_COUNT.set(len(new))



from typing import List

from pygrocy import Grocy
from pygrocy.grocy_api_client import ShoppingListItem

from grocy_telegram_bot.monitoring.watcher import GrocyEntityWatcher
from grocy_telegram_bot.stats import SHOPPING_LIST_WATCHER_TIME


class ShoppingListWatcher(GrocyEntityWatcher):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(grocy, on_update_listener, interval)

    def _fetch_data(self) -> List[ShoppingListItem]:
        return self.grocy.shopping_list(True)

    @SHOPPING_LIST_WATCHER_TIME.time()
    def _run(self):
        super()._run()



from prometheus_client import Summary, Gauge
from prometheus_client.metrics import MetricWrapperBase

from grocy_telegram_bot.const import *

COMMAND_TIME = Summary('command_processing_seconds', 'Time spent in a command handler', ['command'])
COMMAND_TIME_START = COMMAND_TIME.labels(command=COMMAND_START)
COMMAND_TIME_CHORES = COMMAND_TIME.labels(command=COMMAND_CHORES)
COMMAND_TIME_INVENTORY = COMMAND_TIME.labels(command=COMMAND_INVENTORY)
COMMAND_TIME_SHOPPING = COMMAND_TIME.labels(command=COMMAND_SHOPPING)
COMMAND_TIME_SHOPPING_LIST = COMMAND_TIME.labels(command=COMMAND_SHOPPING_LIST)
COMMAND_TIME_SHOPPING_LIST_ADD = COMMAND_TIME.labels(command=COMMAND_SHOPPING_LIST_ADD)

PRODUCT_INVENTORY_COUNT = Gauge(
    'product_inventory_count',
    'Number of inventory items per product name',
    ['product_name']
)

EXPIRED_PRODUCTS_COUNT = Gauge(
    'expired_products_count',
    'Number of expired products in inventory'
)

PRODUCTS_BELOW_MINIMUM_STOCK_COUNT = Gauge(
    'products_below_minimum_stock_count',
    'Number of items a product is below its minimum stock',
    ['product_name']
)

CHORES_COUNT = Gauge(
    'chores_count',
    'Number of overdue chores',
    ['type']
)

OVERDUE_CHORES_COUNT = CHORES_COUNT.labels(type="overdue")
TOTAL_CHORES_COUNT = CHORES_COUNT.labels(type="total")

SHOPPING_LIST_ITEM_COUNT = Gauge(
    'shopping_list_item_count',
    'Number items in a shopping list',
    ['name']
)

TASK_COUNT = Gauge(
    'task_count',
    'Number tasks',
)

WATCHER_TIME = Summary('watcher_processing_seconds', 'Time spent in a Watcher run', ['type'])

CHORE_WATCHER_TIME = WATCHER_TIME.labels(type="chore")
STOCK_WATCHER_TIME = WATCHER_TIME.labels(type="stock")
VOLATILE_STOCK_WATCHER_TIME = WATCHER_TIME.labels(type="volatile_stock")
SHOPPING_LIST_WATCHER_TIME = WATCHER_TIME.labels(type="shopping_list")
TASK_WATCHER_TIME = WATCHER_TIME.labels(type="task")


def get_metrics() -> []:
    entries = set()
    for name, obj in globals().items():
        if isinstance(obj, MetricWrapperBase):
            entries.add(obj)

    return list(entries)


def format_metrics() -> str:
    def format_sample(sample):
        result = "  "
        if len(sample[0]) > 0:
            result += str(sample[0])
        if len(sample[1]) > 0:
            result += str(sample[1])

        if len(result) > 0:
            result += " "
        result += str(sample[2])

        return result

    def format_samples(samples):
        return "\n".join(list(map(format_sample, samples)))

    def format_metric(metric):
        name = metric._name
        samples = list(metric._samples())
        samples_text = format_samples(samples)

        return "{}:\n{}".format(name, samples_text)

    return "\n\n".join(map(format_metric, get_metrics()))



from telegram import Update, ParseMode
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_HELP
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.util import send_message


class HelpCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_HELP,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self.help_command_callback),
        ]

    @command(
        name=COMMAND_HELP,
        description="List commands supported by this bot.",
        permissions=CONFIG_ADMINS,
    )
    def help_command_callback(self, update: Update, context: CallbackContext):
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        from telegram_click import generate_command_list
        text = generate_command_list(update, context)
        send_message(bot, chat_id, text,
                     parse_mode=ParseMode.MARKDOWN,
                     reply_to=message.message_id)



from grocy_telegram_bot.telegram_util import ShoppingListItemButtonCallbackData
from tests import TestBase


class MinifiableDataTest(TestBase):

    def test_minifiable_data(self):
        callback_data = ShoppingListItemButtonCallbackData(
            shopping_list_item_id=123,
            button_click_count=0,
            shopping_list_amount=3
        )

        minified = callback_data.minify()
        self.assertEqual(minified, '{"c":"s","s":123,"b":0,"sh":3}')

        parsed_callback_data = ShoppingListItemButtonCallbackData.parse(minified)
        self.assertEqual(list(vars(callback_data).values()), list(vars(parsed_callback_data).values()))



from typing import List

from telegram.ext import Updater

from grocy_telegram_bot.util import send_message


class Notifier:

    def __init__(self, updater: Updater, chat_ids: List[str]):
        self._chat_ids = chat_ids
        self._updater = updater

    def notify(self, message: str):
        """
        Send notification to all enabled chats
        :param message: the message to send
        """
        for chat_id in self._chat_ids:
            send_message(self._updater.bot, chat_id, message)



from datetime import datetime

REQUESTS_TIMEOUT = (5, 5)
TELEGRAM_CAPTION_LENGTH_LIMIT = 200

NEVER_EXPIRES_DATE = datetime(year=2999, month=12, day=31).date()

# Commands
COMMAND_START = "start"
COMMAND_CHAT_ID = "chat_id"

COMMAND_INVENTORY = ["inventory", "i"]
COMMAND_INVENTORY_ADD = ["inventory_add", "ia"]
COMMAND_INVENTORY_REMOVE = ["inventory_remove", "ir"]
COMMAND_CHORES = ["chores", "ch"]
COMMAND_SHOPPING = ["shopping", "s"]
COMMAND_SHOPPING_LIST = ["shopping_list", "sl"]
COMMAND_SHOPPING_LIST_ADD = ["shopping_list_add", "sla"]

COMMAND_STATS = 'stats'

COMMAND_HELP = ['help', 'h']
COMMAND_VERSION = ['version', 'v']
COMMAND_CONFIG = ['config', 'c']

CANCEL_KEYBOARD_COMMAND = "/cancel_keyboard"



import json
from typing import Dict

from grocy_telegram_bot.const import COMMAND_SHOPPING


class MinifiableData:

    def __init__(self, *args):
        self._minified_key_to_name = {}
        self._name_to_minified_key = {}

    @classmethod
    def parse(cls, text: str):
        minified_dict = json.loads(text)
        # create with random order first
        # noinspection PyArgumentList
        instance = cls(*minified_dict.values())
        # then generate minified key mapping
        instance._generate_minified_keys()
        # and reassign all properties
        for minified_key, name in instance._minified_key_to_name.items():
            vars(instance)[name] = minified_dict[minified_key]

        return instance

    def _get_properties(self):
        return dict(filter(lambda x: not x[0].startswith('_'), vars(self).items()))

    def minify(self) -> str:
        """
        Creates a minified json version of this object
        :return: minified json
        """
        self._generate_minified_keys()

        properties = self._get_properties()
        minified_dict = {}
        for name, minified_key in self._name_to_minified_key.items():
            minified_dict[minified_key] = properties[name]

        minified = self._json_minified(minified_dict)
        if len(minified) > 64:
            raise ValueError(f"Cant minify object, because it would still be to long: {self}")
        return minified

    @staticmethod
    def _json_minified(data: Dict) -> str:
        """
        Convert a dictionary to a minified json representation
        :param data: the data to convert
        :return: minified json string
        """
        return json.dumps(data, indent=None, separators=(',', ':'))

    def _generate_minified_keys(self):
        self._minified_key_to_name.clear()
        self._name_to_minified_key.clear()

        properties = self._get_properties()
        for name in properties.keys():
            max_len = len(name)
            length = 1
            while length <= max_len:
                minified_key = name[0:length]
                if minified_key in self._minified_key_to_name:
                    length += 1
                else:
                    self._minified_key_to_name[minified_key] = name
                    self._name_to_minified_key[name] = minified_key
                    break


class CallbackData(MinifiableData):
    command_id: str

    def __init__(self, *args):
        super().__init__(args)
        self.command_id = self.__class__.command_id


class ShoppingListItemButtonCallbackData(CallbackData):
    command_id: str = COMMAND_SHOPPING[1]

    def __init__(self, shopping_list_item_id: int, button_click_count: int, shopping_list_amount: int,
                 *args):
        super().__init__(args)
        self.shopping_list_item_id = shopping_list_item_id
        self.button_click_count = button_click_count
        self.shopping_list_amount = shopping_list_amount



from datetime import datetime, timedelta

from pygrocy.grocy import Product
from telegram import ParseMode, Update, ReplyKeyboardRemove
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.argument import Flag, Argument
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_INVENTORY, COMMAND_INVENTORY_ADD, NEVER_EXPIRES_DATE, \
    COMMAND_INVENTORY_REMOVE
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.stats import COMMAND_TIME_INVENTORY
from grocy_telegram_bot.util import send_message, product_to_str, timing


class InventoryCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_INVENTORY,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._inventory_callback),
            CommandHandler(COMMAND_INVENTORY_ADD,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._inventory_add_callback),
            CommandHandler(COMMAND_INVENTORY_REMOVE,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._inventory_remove_callback),
        ]

    @command(
        name=COMMAND_INVENTORY,
        description="List product inventory.",
        arguments=[
            Flag(name=["missing", "m"], description="Show missing products"),
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_INVENTORY.time()
    def _inventory_callback(self, update: Update, context: CallbackContext, missing: bool) -> None:
        """
        Show a list of all products in the inventory
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id

        products = self._grocy.get_all_products()
        if missing:
            products = list(filter(lambda x: x.amount == 0, products))

        products = sorted(products, key=lambda x: x.name.lower())

        item_texts = list(list(map(product_to_str, products)))
        text = "\n".join([
            "*=> Inventory <=*",
            *item_texts,
        ]).strip()

        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN)

    @command(
        name=COMMAND_INVENTORY_REMOVE,
        description="Remove a product from inventory.",
        arguments=[
            Argument(name=["name"], description="Product name", example="Banana"),
            Argument(name=["amount"], description="Product amount", type=int, example="2",
                     validator=lambda x: x > 0, optional=True, default=1),
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_INVENTORY.time()
    def _inventory_remove_callback(self, update: Update, context: CallbackContext,
                                   name: str, amount: int) -> None:
        """
        Add a product to the inventory
        :param update: the chat update object
        :param context: telegram context
        """
        products = self._grocy.get_all_products()
        self._reply_keyboard_handler.await_user_selection(
            update, context, name, choices=products, key=lambda x: x.name,
            callback=self._remove_product_keyboard_response_callback,
            callback_data={
                "product_name": name,
                "amount": amount,
            }
        )

    @command(
        name=COMMAND_INVENTORY_ADD,
        description="Add a product to inventory.",
        arguments=[
            Argument(name=["name"], description="Product name", example="Banana"),
            Argument(name=["amount"], description="Product amount", type=int, example="2",
                     validator=lambda x: x > 0, optional=True, default=1),
            Argument(name=["exp"], description="Expiration date or duration", example="20.01.2020",
                     optional=True, default="Never"),
            Argument(name=["price"], description="Product price", type=float, example="2.80",
                     validator=lambda x: x > 0, optional=True),
        ],
        permissions=CONFIG_ADMINS
    )
    @COMMAND_TIME_INVENTORY.time()
    def _inventory_add_callback(self, update: Update, context: CallbackContext,
                                name: str, amount: int, exp: str, price: float or None) -> None:
        """
        Add a product to the inventory
        :param update: the chat update object
        :param context: telegram context
        """
        # parse the expiration date input
        if exp.casefold() == "never".casefold():
            exp = NEVER_EXPIRES_DATE
        else:
            try:
                from dateutil import parser
                exp = parser.parse(exp)
            except:
                from pytimeparse import parse
                parsed = parse(exp)
                if parsed is None:
                    raise ValueError("Cannot parse the given time format: {}".format(exp))
                exp = datetime.now() + timedelta(seconds=parsed)

        products = self._grocy.get_all_products()
        self._reply_keyboard_handler.await_user_selection(
            update, context, name, choices=products, key=lambda x: x.name,
            callback=self._add_product_keyboard_response_callback,
            callback_data={
                "product_name": name,
                "amount": amount,
                "exp": exp,
                "price": price
            }
        )

    def _add_product_keyboard_response_callback(self, update: Update, context: CallbackContext,
                                                product: Product, data: dict):
        """
        Called when the user has selected a product to add to the inventory
        :param update: the chat update object
        :param context: telegram context
        :param product: the selected product
        :param data: callback data
        """
        amount = data["amount"]
        exp = data["exp"]
        price = data["price"]

        self._inventory_add_execute(update, context, product, amount, exp, price)

    @timing
    def _remove_product_keyboard_response_callback(self, update: Update, context: CallbackContext,
                                                   product: Product, data: dict):
        """
        Called when the user has selected a product to remove from to the inventory
        :param update: the chat update object
        :param context: telegram context
        :param product: the selected product
        :param data: callback data
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id
        amount = data["amount"]

        self._grocy.add_product(product_id=product.id, amount=-amount, price=None)

        text = "Removed {}x {}".format(amount, product.name)
        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN, reply_to=message_id,
                     menu=ReplyKeyboardRemove(selective=True))

    def _inventory_add_execute(self, update: Update, context: CallbackContext, product: Product, amount: int,
                               exp: datetime, price: float):
        """
        Adds a product to the inventory
        :param update: the chat update object
        :param context: telegram context
        :param product: product entity
        :param amount: amount
        :param exp: expiration date
        :param price: price
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id
        self._grocy.add_product(product_id=product.id, amount=amount, price=price, best_before_date=exp)

        text = "Added {}x {} (Exp: {}, Price: {})".format(
            amount, product.name, "Never" if exp == NEVER_EXPIRES_DATE else exp, price)
        send_message(bot, chat_id, text, parse_mode=ParseMode.MARKDOWN, reply_to=message_id,
                     menu=ReplyKeyboardRemove(selective=True))



from typing import List

from telegram.ext import Handler

from grocy_telegram_bot.bot import ReplyKeyboardHandler, InlineKeyboardHandler
from grocy_telegram_bot.cache import GrocyCached
from grocy_telegram_bot.config import Config


class GrocyCommandHandler:

    def __init__(self, config: Config, grocy: GrocyCached, reply_keyboard_handler: ReplyKeyboardHandler,
                 inline_keyboard_handler: InlineKeyboardHandler):
        self._config = config
        self._grocy = grocy
        self._reply_keyboard_handler = reply_keyboard_handler
        self._inline_keyboard_handler = inline_keyboard_handler

    def command_handlers(self) -> List[Handler]:
        """
        :return: a list of all command handlers
        """
        raise NotImplementedError()



import logging
from typing import List

from expiringdict import ExpiringDict
from pygrocy import Grocy
from pygrocy.grocy import Product

from grocy_telegram_bot.config import Config
from grocy_telegram_bot.util import timing

LOGGER = logging.getLogger(__name__)

CONFIG = Config()
# CACHE = LruCache(expires=CONFIG.GROCY_CACHE_DURATION.value.total_seconds(), concurrent=True)
CACHE = ExpiringDict(max_len=100, max_age_seconds=CONFIG.GROCY_CACHE_DURATION.value.total_seconds())

FUNCTIONS_TO_CACHE = [
    "Grocy.stock",
    "Grocy.volatile_stock",
    "Grocy.chore",
    "Grocy.chores",
    "Grocy.product",
    "Grocy.shopping_list",
    "Grocy.product_groups",
    "Grocy.get_userfields",
    "Grocy.get_last_db_changed",
    "Grocy.expired_products",
    "Grocy.expiring_products",
    "Grocy.missing_products",
    "GrocyCached.get_all_products",
]


def cache_decorator(func: classmethod):
    """
    Decorator to cache the result of a function call
    :param func: the function to wrap
    :return: wrapped function
    """

    def wrapper(*args, **kwargs):
        if not func.__qualname__.startswith("Grocy"):
            LOGGER.warning("Using cache on other object than Grocy, ignoring cache")
            return func(*args, **kwargs)

        if func.__qualname__ not in FUNCTIONS_TO_CACHE:
            LOGGER.debug(f"Clearing cache because of non-whitelisted function call: {func.__qualname__}")
            # clear existing cache since the data will probably change
            CACHE.clear()
            # don't cache if not whitelisted
            return func(*args, **kwargs)

        key = f"{func.__qualname__}_{args}_{kwargs}"

        if key in CACHE:
            return CACHE[key]

        response = func(*args, **kwargs)
        LOGGER.debug(f"Caching function response: {key}")
        CACHE[key] = response
        return response

    return wrapper


class GrocyCached(Grocy):

    def __getattribute__(self, name):
        ret = super(Grocy, self).__getattribute__(name)
        if callable(ret):
            return cache_decorator(ret)
        else:
            return ret

    @timing
    def get_all_products(self) -> List[Product]:
        """
        Get a list of all products
        :return: produtcs
        """
        stock = self.stock(True)
        ex_stock = self.expiring_products(True)
        ex2_stock = self.expired_products(True)
        # TODO: add when fixed upstream
        # m_stock = self.missing_products(True)
        return stock + ex_stock + ex2_stock



from typing import List

from pygrocy import Grocy
from pygrocy.grocy import Product

from grocy_telegram_bot.monitoring.watcher import GrocyEntityWatcher
from grocy_telegram_bot.stats import VOLATILE_STOCK_WATCHER_TIME, STOCK_WATCHER_TIME


class StockWatcher(GrocyEntityWatcher):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(grocy, on_update_listener, interval)

    def _fetch_data(self) -> List[Product]:
        return self.grocy.stock(True)

    @STOCK_WATCHER_TIME.time()
    def _run(self):
        super()._run()


class VolatileStockWatcher(GrocyEntityWatcher):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(grocy, on_update_listener, interval)

    def _fetch_data(self) -> List[Product]:
        # TODO: enable this when fixed in pygrocy
        # missing = self.grocy.missing_products(True)
        expiring = self.grocy.expiring_products(True)
        expired = self.grocy.expired_products(True)
        # return missing + expiring + expired
        return expiring + expired

    @VOLATILE_STOCK_WATCHER_TIME.time()
    def _run(self):
        super()._run()



import logging
import threading
from typing import List

from pygrocy import Grocy

LOGGER = logging.getLogger(__name__)


class RegularIntervalWorker:
    """
    Base class for a worker that executes a specific task in a regular interval.
    """

    def __init__(self, interval: float):
        self._interval = interval
        self._timer = None

    def start(self):
        """
        Starts the worker
        """
        if self._timer is None:
            LOGGER.debug(f"Starting worker: {self.__class__.__name__}")
            self._schedule_next_run(0)
        else:
            LOGGER.debug("Already running, ignoring start() call")

    def stop(self):
        """
        Stops the worker
        """
        if self._timer is not None:
            self._timer.cancel()
        self._timer = None

    def _schedule_next_run(self, interval: int = None):
        """
        Schedules the next run
        """
        if self._timer is not None:
            self._timer.cancel()
        interval = interval if interval is not None else self._interval
        self._timer = threading.Timer(interval, self._worker_job)
        self._timer.start()

    def _worker_job(self):
        """
        The regularly executed task. Override this method.
        """
        try:
            self._run()
        except Exception as e:
            LOGGER.error(e, exc_info=True)
        finally:
            self._schedule_next_run()

    def _run(self):
        """
        The regularly executed task. Override this method.
        """
        raise NotImplementedError()


class GrocyEntityWatcher(RegularIntervalWorker):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(interval)
        self.grocy = grocy
        self.on_update_listener = on_update_listener
        self.data = None

    def _fetch_data(self) -> List:
        """
        Fetch the data this watcher should watch.
        """
        raise NotImplementedError()

    def _run(self):
        data = self._fetch_data()

        try:
            self.on_update_listener(self.data, data)
        finally:
            self.data = data



from telegram import Update
from telegram.ext import Filters, CommandHandler, CallbackContext
from telegram_click.decorator import command

from grocy_telegram_bot.commands import GrocyCommandHandler
from grocy_telegram_bot.const import COMMAND_VERSION
from grocy_telegram_bot.permissions import CONFIG_ADMINS
from grocy_telegram_bot.util import send_message


class VersionCommandHandler(GrocyCommandHandler):

    def command_handlers(self):
        return [
            CommandHandler(COMMAND_VERSION,
                           filters=(~ Filters.reply) & (~ Filters.forwarded),
                           callback=self._version_command_callback),
        ]

    @command(
        name=COMMAND_VERSION,
        description="Show the version of this bot.",
        permissions=CONFIG_ADMINS
    )
    def _version_command_callback(self, update: Update, context: CallbackContext) -> None:
        """
        /stats command handler
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        from grocy_telegram_bot import __version__
        text = "{}".format(__version__)
        send_message(bot, chat_id, text, reply_to=message.message_id)



from typing import List

from pygrocy import Grocy
from pygrocy.grocy import Chore

from grocy_telegram_bot.monitoring.watcher import GrocyEntityWatcher
from grocy_telegram_bot.stats import CHORE_WATCHER_TIME


class ChoreWatcher(GrocyEntityWatcher):

    def __init__(self, grocy: Grocy, on_update_listener, interval: float):
        super().__init__(grocy, on_update_listener, interval)

    def _fetch_data(self) -> List[Chore]:
        return self.grocy.chores(True)

    @CHORE_WATCHER_TIME.time()
    def _run(self):
        super()._run()



import logging
import re

from container_app_conf import ConfigBase
from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.list import ListConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry
from container_app_conf.source.env_source import EnvSource
from container_app_conf.source.toml_source import TomlSource
from container_app_conf.source.yaml_source import YamlSource
from py_range_parse import Range

NODE_MAIN = "grocy_telegram_bot"

NODE_NOTIFICATION = "notification"
NODE_TELEGRAM = "telegram"

NODE_GROCY = "grocy"
NODE_HOST = "host"
NODE_API_KEY = "api_key"

NODE_STATS = "stats"
NODE_ENABLED = "enabled"
NODE_PORT = "port"


class Config(ConfigBase):

    def __new__(cls, *args, **kwargs):
        yaml_source = YamlSource("grocy_telegram_bot")
        toml_source = TomlSource("grocy_telegram_bot")
        data_sources = [
            EnvSource(),
            yaml_source,
            toml_source
        ]
        return super(Config, cls).__new__(cls, data_sources=data_sources)

    LOG_LEVEL = StringConfigEntry(
        description="Log level",
        key_path=[
            NODE_MAIN,
            "log_level"
        ],
        regex=re.compile(f" {'|'.join(logging._nameToLevel.keys())}", flags=re.IGNORECASE),
        default="WARNING",
    )

    LOCALE = StringConfigEntry(
        description="Bot Locale",
        key_path=[
            NODE_MAIN,
            "locale"
        ],
        default="en",
    )

    TELEGRAM_BOT_TOKEN = StringConfigEntry(
        description="The telegram bot token to use",
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            "bot_token"
        ],
        example="123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11",
        secret=True)

    TELEGRAM_ADMIN_USERNAMES = ListConfigEntry(
        item_type=StringConfigEntry,
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            "admin_usernames"
        ],
        required=True,
        example=[
            "myadminuser",
            "myotheradminuser"
        ]
    )

    GROCY_CACHE_DURATION = TimeDeltaConfigEntry(
        description="Duration to cache Grocy REST api call responses",
        key_path=[
            NODE_MAIN,
            NODE_GROCY,
            "cache_duration"
        ],
        required=True,
        default="60s",
    )

    NOTIFICATION_CHAT_IDS = ListConfigEntry(
        item_type=StringConfigEntry,
        key_path=[
            NODE_MAIN,
            NODE_NOTIFICATION,
            "chat_ids"
        ],
        default=[]
    )

    GROCY_HOST = StringConfigEntry(
        description="Hostname of the Grocy instance",
        key_path=[
            NODE_MAIN,
            NODE_GROCY,
            NODE_HOST
        ],
        required=True,
        default="127.0.0.1"
    )

    GROCY_PORT = IntConfigEntry(
        description="Port of the Grocy REST api",
        key_path=[
            NODE_MAIN,
            NODE_GROCY,
            NODE_PORT
        ],
        range=Range(1, 65535),
        default=80
    )

    GROCY_API_KEY = StringConfigEntry(
        description="Grocy API Key used for REST authentication",
        key_path=[
            NODE_MAIN,
            NODE_GROCY,
            NODE_API_KEY
        ],
        required=True,
        example="abcdefgh12345678",
        secret=True
    )

    STATS_ENABLED = BoolConfigEntry(
        description="Whether to enable prometheus statistics or not.",
        key_path=[
            NODE_MAIN,
            NODE_STATS,
            NODE_ENABLED
        ],
        default=True
    )

    STATS_PORT = IntConfigEntry(
        description="The port to expose statistics on.",
        key_path=[
            NODE_MAIN,
            NODE_STATS,
            NODE_PORT
        ],
        default=8000
    )
